/*
 * Core OCR algorithm (first exact bitmap match)
 */
/*
 *  This is part of ypp-sc-tools, a set of third-party tools for assisting
 *  players of Yohoho Puzzle Pirates.
 * 
 *  Copyright (C) 2009 Ian Jackson <ijackson@chiark.greenend.org.uk>
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *  Yohoho and Puzzle Pirates are probably trademarks of Three Rings and
 *  are used without permission.  This program is not endorsed or
 *  sponsored by Three Rings.
 */

#include "ocr.h"

typedef struct {
  Pixcol col;
  struct DatabaseNode *then;
} DatabaseLink;

#define MAXGLYPHCHRS 7

typedef struct DatabaseNode {
  char s[MAXGLYPHCHRS+1]; /* null-terminated; "" means no match here */
  int nlinks, alinks;
  unsigned endsword:1;
  DatabaseLink *links;
} DatabaseNode;

static const char *context_names[]= {
  "Lower",
  "Upper",
  "Digit"
};
struct OcrCellTypeInfo {
  /* bitmaps of indices into context_names: */
  unsigned initial, nextword, midword;
  int space_spaces;
  const char *name;
};
const struct OcrCellTypeInfo ocr_celltype_number= {
  4,4,4,
  .space_spaces= 5,
  .name= "number"
};
const struct OcrCellTypeInfo ocr_celltype_text= {
  .initial=2, /* Uppercase */
  .nextword=3, /* Either */
  .midword=1, /* Lower only */
  .space_spaces= 4,
  .name= "text"
};


#define NCONTEXTS (sizeof(context_names)/sizeof(context_names[0]))

struct OcrReader {
  int h;
  DatabaseNode contexts[NCONTEXTS];
  OcrResultGlyph *results;
  int aresults, nresults;
};

static FILE *resolver;
static pid_t resolver_pid;
static int resolver_done;

DEBUG_DEFINE_DEBUGF(ocr)

#define dbassert(x)							\
  ((x) ? (void)0 :							\
     fatal("Error in character set database.\n"				\
	   " Requirement not met: %s:%d: %s", __FILE__,__LINE__, #x))

static void fgetsline(FILE *f, char *lbuf, size_t lbufsz) {
  errno=0;
  char *s= fgets(lbuf,lbufsz,f);
  sysassert(!ferror(f));
  dbassert(!feof(f));
  assert(s);
  int l= strlen(lbuf);
  dbassert(l>0);  dbassert(lbuf[--l]='\n');
  lbuf[l]= 0;
}
#define FGETSLINE(f,buf) (fgetsline(f,buf,sizeof(buf)))

static void cleardb_node(DatabaseNode *n) {
  int i;
  n->s[0]= 0;
  for (i=0; i<n->nlinks; i++)
    cleardb_node(n->links[i].then);
}

static void readdb(OcrReader *rd) {
  int nchrs;
  DatabaseNode *current, *additional;
  char chrs[MAXGLYPHCHRS+1];
  Pixcol cv;
  int r,j,ctxi;
  int h, endsword;
  char lbuf[100];
  FILE *db;

  for (ctxi=0; ctxi<NCONTEXTS; ctxi++)
    cleardb_node(&rd->contexts[ctxi]);

  char *dbfname=0;
  asprintf(&dbfname,"%s/charset-%d.txt",get_vardir(),rd->h);
  sysassert(dbfname);
  
  db= fopen(dbfname,"r");
  free(dbfname);
  if (!db) {
    sysassert(errno==ENOENT);
    return;
  }

  FGETSLINE(db,lbuf);
  dbassert(!strcmp(lbuf,"# ypp-sc-tools pctb font v1"));

  r= fscanf(db, "%d", &h);
  dbassert(r==1);
  dbassert(h==rd->h);

  for (;;) {
    FGETSLINE(db,lbuf);
    if (!lbuf || lbuf[0]=='#') continue;
    if (!strcmp(lbuf,".")) break;

    for (ctxi=0; ctxi<NCONTEXTS; ctxi++)
      if (!strcmp(lbuf,context_names[ctxi]))
	goto found_ctx;
    /* not found, just skip */
    for (;;) { FGETSLINE(db,lbuf); if (!lbuf[0]) break; }
    continue;

  found_ctx:
    for (nchrs=0;;) {
      int c= fgetc(db); sysassert(!ferror(db)); dbassert(c!=EOF);
      if (c=='\n') { dbassert(nchrs); break; }
      dbassert(nchrs<MAXGLYPHCHRS);
      if (c=='\\') {
	unsigned cr;
	c= fgetc(db); sysassert(!ferror(db)); dbassert(c=='x');
	r= fscanf(db, "%2x", &cr); sysassert(!ferror(db)); dbassert(r==1);
	assert(cr>0 && cr<=255);
	c= cr;
      }
      chrs[nchrs++]= c;
    }
    endsword= 0;
    if (nchrs>1 && chrs[nchrs-1]==' ') {
      endsword= 1;
      nchrs--;
    }
    chrs[nchrs]= 0;

    current= &rd->contexts[ctxi];
    for (;;) {
      FGETSLINE(db,lbuf);
      if (!lbuf[0]) { dbassert(current != &rd->contexts[ctxi]); break; }
      char *ep;
      cv= strtoul(lbuf,&ep,16);  dbassert(!*ep);
      dbassert(!(cv & ~((1UL << rd->h)-1)));
      
      for (j=0; j<current->nlinks; j++)
	if (current->links[j].col == cv) {
	  current= current->links[j].then;
	  goto found_link;
	}

      additional= mmalloc(sizeof(*additional));
      additional->s[0]= 0;
      additional->nlinks= additional->alinks= 0;
      additional->links= 0;
      if (current->nlinks==current->alinks) {
	current->alinks++;
	current->alinks<<=1;
	current->links= mrealloc(current->links,
				 sizeof(*current->links) * current->alinks);
      }
      current->links[current->nlinks].col= cv;
      current->links[current->nlinks].then= additional;
      current->nlinks++;
      current= additional;

    found_link:;
    }

    dbassert(!current->s[0]);
    strcpy(current->s, chrs);
    current->endsword= endsword;
  }
  sysassert(!ferror(db));
  sysassert(!fclose(db));
}

static void cu_pr_ctxmap(unsigned ctxmap) {
  fprintf(resolver,"{");
  const char *spc="";
  int ctxi;
  for (ctxi=0; ctxi<NCONTEXTS; ctxi++) {
    if (!(ctxmap & (1u << ctxi))) continue;
    fprintf(resolver,"%s%s",spc,context_names[ctxi]);
    spc=" ";
  }
  fprintf(resolver,"}");
}

static void callout_unknown(OcrReader *rd, int w, Pixcol cols[],
			    int unk_l, int unk_r, unsigned unk_ctxmap) {
  int jobpipe[2],donepipe[2], c,i, x,y;
  const OcrResultGlyph *s;
  const char *p;
  char cb;
  Pixcol pv;
  
  if (!resolver) {
    sysassert(! pipe(jobpipe) );
    sysassert(! pipe(donepipe) );
    resolver_pid= fork();
    sysassert(resolver_pid!=-1);
    if (!resolver_pid) {
      sysassert( dup2(jobpipe[0],0) ==0 );
      sysassert(! close(jobpipe[1]) );
      sysassert(! close(donepipe[0]) );
      /* we know donepipe[1] is >= 4 and we have dealt with all the others
       * so we aren't in any danger of overwriting some other fd 4: */
      sysassert( dup2(donepipe[1],4) ==4 );
      execlp("./yppsc-ocr-resolver", "yppsc-ocr-resolver",
	     DEBUGP(callout) ? "--debug" : "--noop-arg",
	     "--automatic-1",
	     (char*)0);
      sysassert(!"execlp ocr-resolver failed");
    }
    sysassert(! close(jobpipe[0]) );
    sysassert(! close(donepipe[1]) );
    resolver= fdopen(jobpipe[1],"w"); sysassert(resolver);
    resolver_done= donepipe[0];
  }
  fprintf(resolver,"%d %d ",unk_l,unk_r);
  cu_pr_ctxmap(unk_ctxmap);
  for (i=0, s=rd->results; i<rd->nresults; i++, s++) {
    if (!strcmp(s->s," ")) continue;
    fprintf(resolver," %d %d ",s->l,s->r);
    cu_pr_ctxmap(s->ctxmap);
    fprintf(resolver," ");
    for (p=s->s; (c= *p); p++) {
      if (c=='\\') fprintf(resolver,"\\%c",c);
      else if (c>=33 && c<=126) fputc(c,resolver);
      else fprintf(resolver,"\\x%02x",(unsigned char)c);
    }
  }
  fputc('\n',resolver);

  fprintf(resolver,
	  "/* XPM */\n"
	  "static char *t[] = {\n"
	  "/* columns rows colors chars-per-pixel */\n"
	  "\"%d %d 2 1\",\n"
	  "\"  c black\",\n"
	  "\"o c white\",\n",
	  w,rd->h);
  for (y=0, pv=1; y<rd->h; y++, pv<<=1) {
    fputc('"',resolver);
    for (x=0; x<w; x++)
      fputc(cols[x] & pv ? 'o' : ' ', resolver);
    fputs("\",\n",resolver);
  }
  fputs("};\n",resolver);
  sysassert(!ferror(resolver));
  sysassert(!fflush(resolver));

  sysassert(resolver);

  int r;
  for (;;) {
    r= read(resolver_done,&cb,1);
    if (r==-1) { sysassert(errno==EINTR); continue; }
    break;
  }

  if (r==0) {
    waitpid_check_exitstatus(resolver_pid, "character resolver");
    fclose(resolver);
    close(resolver_done);
    resolver= 0;
  } else {
    assert(r==1);
    sysassert(cb==0);
  }

  readdb(rd);
}

static void add_result(OcrReader *rd, const char *s, int l, int r,
		       unsigned ctxmap) {
  if (rd->nresults >= rd->aresults) {
    rd->aresults++; rd->aresults<<=1;
    rd->results= mrealloc(rd->results, sizeof(*rd->results)*rd->aresults);
  }
  rd->results[rd->nresults].s= s;
  rd->results[rd->nresults].l= l;
  rd->results[rd->nresults].r= r;
  rd->results[rd->nresults].ctxmap= ctxmap;
  rd->nresults++;
}


const char *ocr_celltype_name(OcrCellType ct) { return ct->name; }

OcrResultGlyph *ocr(OcrReader *rd, OcrCellType ct, int w, Pixcol cols[]) {
  int nspaces;
  unsigned ctxmap;
  int ctxi, i, x;

 restart:

  nspaces=- w;
  ctxmap= ct->initial;
  rd->nresults=0;
  debugf("OCR h=%d w=%d",rd->h,w);
  for (x=0; x<w; x++) debugf(" %"PSPIXCOL(PRIx),cols[x]);
  debugf("\n");
  debug_flush();

  x=0;
  for (;;) {
    debug_flush();
    /* skip spaces */
    if (x>=w)
      break;

    if (!cols[x]) {
      nspaces++;
      x++;
      if (nspaces == ct->space_spaces) {
	debugf("OCR  x=%x nspaces=%d space\n",x,nspaces);
	ctxmap= ct->nextword;
      }
      continue;
    }

    /* something here, so we need to add the spaces */
    if (nspaces >= ct->space_spaces)
      add_result(rd," ",x-nspaces,x+1,0);
    nspaces=0;

    /* find character */
    int lx=x;

    DatabaseNode *uniquematch= 0;
    int uniquematch_rx=-1;
    
    debugf("OCR  lx=%d ctxmap=%x  ",lx,ctxmap);

    for (ctxi=0; ctxi<NCONTEXTS; ctxi++) {
      DatabaseNode *current= &rd->contexts[ctxi];;
      DatabaseNode *bestmatch= 0;
      int bestmatch_rx=-1;

      x= lx;
      if (!(ctxmap & (1u << ctxi))) continue;
      debugf(" || %s",context_names[ctxi]);

      for (;;) {
	debug_flush();
	debugf(" | x=%d",x);
	if (x>w) break;
	Pixcol cv= cols[x];
	debugf(" cv=%"PSPIXCOL(PRIx),cv);
	for (i=0; i<current->nlinks; i++)
	  if (current->links[i].col == cv)
	    goto found;
	/* not found */
	debugf(" ?");
	break;

      found:
	current= current->links[i].then;
	if (current->s[0]) {
	  debugf(" \"%s\"%s",current->s,current->endsword?"_":"");
	  bestmatch= current;
	  bestmatch_rx= x;
	} else {
	  debugf(" ...");
	}

	x++;
      }
      
      if (bestmatch) {
	if (uniquematch && strcmp(bestmatch->s, uniquematch->s)) {
	  debugf( " ambiguous");
	  uniquematch= 0;
	  break;
	}
	uniquematch= bestmatch;
	uniquematch_rx= bestmatch_rx;
      }
    }

    if (uniquematch) {
      debugf(" || YES");
      add_result(rd, uniquematch->s, lx, uniquematch_rx, ctxmap);
      x= uniquematch_rx+1;
      if (uniquematch->s[0]) ctxmap= ct->midword;
      else debugf(" (empty)");
      if (uniquematch->endsword) {
	nspaces= ct->space_spaces;
	debugf("_");
	ctxmap= ct->nextword;
      }
      debugf("\n");
    } else {
      int rx;
      debugf(" || UNKNOWN");
      for (rx=lx; rx<w && cols[rx]; rx++);
      debugf(" x=%d ctxmap=%x %d..%d\n",x, ctxmap, lx,rx);
      debug_flush();
      callout_unknown(rd, w,cols, lx,rx-1, ctxmap);
      goto restart;
    }
  }
  add_result(rd, 0,-1,-1,0);
  debugf("OCR  finished %d glyphs\n",rd->nresults);
  debug_flush();
  return rd->results;
}

OcrReader *ocr_init(int h) {
  OcrReader *rd;

  rd= mmalloc(sizeof(*rd));
  memset(rd,0,sizeof(*rd));
  rd->h= h;
  readdb(rd);
  return rd;
}
